Object.defineProperty(exports, "__esModule", { value: true });
const db = require("db");
const url = require("url");
const assert = require("assert");
const coroutine = require("coroutine");
const Utils = require("../utils");
function setReactivePool(driver) {
    let pool = false;
    Object.defineProperty(driver.extend_config, 'pool', {
        set(nextVal) {
            if (typeof nextVal === 'string')
                nextVal = Utils.castQueryStringToBoolean(nextVal);
            if (nextVal) {
                pool = Utils.parsePoolConfig(nextVal);
                pool.maxsize = Utils.forceInteger(pool.maxsize, 100);
                pool.timeout = Utils.forceInteger(pool.timeout, 1000);
                Utils.mountPoolToDriver(driver, pool);
            }
            else {
                pool = false;
            }
        },
        get() { return pool; }
    });
}
class Driver {
    constructor(options) {
        this.extend_config = {
            pool: false,
            debug: false
        };
        options = Utils.parseConnectionString(options);
        Object.defineProperty(this, 'config', { get() { return options; } });
        assert.ok(!!this.config.protocol, '[driver.config] invalid protocol');
        // some db has no host
        // assert.ok(!!this.config.host, '[driver.config] invalid host')
        this.type = Utils.filterDriverType(this.config.protocol);
        const extend_config = {};
        Object.defineProperty(this, 'extend_config', { get() { return extend_config; } });
        setReactivePool(this);
        extend_config.pool = options.query.pool;
        extend_config.debug = Utils.castQueryStringToBoolean(options.query.debug);
        Object.defineProperty(this, 'uid', { value: Utils.driverUUid(), writable: false, configurable: false });
    }
    static create(input) {
        const driver = Driver.getDriver(typeof input === 'object' ? input.protocol : input);
        return new driver(input);
    }
    get uri() {
        return url.format(Object.assign(Object.assign({}, this.config), { slashes: this.config.protocol === 'sqlite:' ? false : this.config.slashes, query: this.config.protocol === 'sqlite:' ? {} : this.config.query }));
    }
    get isPool() {
        return !!this.extend_config.pool;
    }
    get isSql() {
        const p = this.config.protocol || '';
        return ((p === 'mysql:')
            || (p.startsWith('sqlite:')));
    }
    get isNoSql() {
        const p = this.config.protocol || '';
        return ((p === 'mongodb:'));
    }
    get isCommand() {
        const p = this.config.protocol || '';
        return ((p === 'mongodb:')
            || (p === 'redis:'));
    }
    /**
     * @override
     */
    reopen() {
        try {
            this.close();
        }
        catch (error) { }
        return this.open();
    }
    /**
     * @override
     */
    open() {
        return this.connection = this.getConnection();
    }
    /**
     * @override
     */
    close() { }
    /**
     * @override
     */
    ping() { }
    /**
     * @override
     */
    getConnection() { return null; }
    connectionPool(callback) {
        if (this.isPool)
            return this.pool((conn) => callback(conn));
        return callback(this.getConnection());
    }
    useTrans(callback) {
        return this.connectionPool((conn) => {
            if (typeof conn.trans === 'function') {
                const waitor = {
                    ev: new (coroutine.Event)(),
                    result: undefined
                };
                conn.trans(() => {
                    waitor.result = callback(conn);
                    waitor.ev.set();
                });
                waitor.ev.wait();
                return waitor.result;
            }
            else {
                return callback(conn);
            }
        });
    }
}
exports.Driver = Driver;
Driver.getDriver = function getDriver(name) {
    switch (name) {
        case 'mysql':
            return MySQLDriver;
        case 'sqlite':
            return SQLiteDriver;
        case 'redis':
            return RedisDriver;
        case 'mongodb':
            return MongoDriver;
        default:
            if (name) {
                const type = Utils.filterDriverType(url.parse(name).protocol);
                if (type !== 'unknown')
                    return getDriver(type);
            }
            return Driver;
    }
};
class SQLDriver extends Driver {
    constructor() {
        super(...arguments);
        this.currentDb = null;
    }
    switchDb(targetDb) { }
    ;
    /**
     * @override
     */
    begin() { }
    /**
     * @override
     */
    commit() { }
    /**
     * @override
     */
    trans(cb) { return true; }
    /**
     * @override
     */
    rollback() { }
    execute(sql) { return; }
}
exports.SQLDriver = SQLDriver;
class MySQLDriver extends SQLDriver {
    constructor(conn) {
        super(conn);
        this.connection = null;
    }
    switchDb(targetDb) {
        this.execute(db.formatMySQL("use `" + db.escape(targetDb) + "`"));
    }
    open() { return super.open(); }
    close() {
        if (this.connection)
            this.connection.close();
    }
    ping() { return; }
    begin() { return this.connection.begin(); }
    commit() { return this.connection.commit(); }
    trans(cb) { return this.connection.trans(cb); }
    rollback() { return this.connection.rollback(); }
    getConnection() { return db.openMySQL(this.uri); }
    execute(sql) {
        if (this.isPool)
            return this.pool(conn => conn.execute(sql));
        if (!this.connection)
            this.open();
        return this.connection.execute(sql);
    }
}
class SQLiteDriver extends SQLDriver {
    constructor(conn) {
        super(conn);
        this.connection = null;
    }
    open() { return super.open(); }
    close() {
        if (this.connection)
            this.connection.close();
    }
    ping() { return; }
    begin() { return this.connection.begin(); }
    commit() { return this.connection.commit(); }
    trans(cb) { return this.connection.trans(cb); }
    rollback() { return this.connection.rollback(); }
    getConnection() { return db.openSQLite(this.uri); }
    execute(sql) {
        if (this.isPool)
            return this.pool(conn => conn.execute(sql));
        if (!this.connection)
            this.open();
        return this.connection.execute(sql);
    }
}
class RedisDriver extends Driver {
    constructor(conn) {
        super(conn);
        this.connection = null;
    }
    open() { return super.open(); }
    close() {
        if (this.connection)
            this.connection.close();
    }
    ping() { }
    command(cmd, ...args) {
        if (this.isPool)
            return this.pool(conn => conn.command(cmd, ...args));
        if (!this.connection)
            this.open();
        return this.connection.command(cmd, ...args);
    }
    commands(cmds, opts) {
        const { parallel = false } = opts || {};
        const keys = Object.keys(cmds);
        if (parallel)
            return coroutine.parallel(keys, (cmd) => {
                return { cmd, result: this.command(cmd, ...Utils.arraify(cmds[cmd])) };
            });
        else
            return Object.keys(cmds).map((cmd) => {
                return { cmd, result: this.command(cmd, ...Utils.arraify(cmds[cmd])) };
            });
    }
    getConnection() { return db.openRedis(this.uri); }
}
class MongoDriver extends Driver {
    constructor(conn) {
        super(conn);
        this.connection = null;
    }
    reopen() {
        try {
            this.close();
        }
        catch (error) { }
        return this.open();
    }
    open() { return super.open(); }
    close() {
        if (this.connection)
            this.connection.close();
    }
    ping() { }
    command(cmd, arg) {
        return this.commands({ [cmd]: arg });
    }
    commands(cmds, opts) {
        if (this.isPool)
            return this.pool(conn => conn.runCommand(cmds));
        if (!this.connection)
            this.open();
        return this.connection.runCommand(cmds);
    }
    getConnection() { return db.openMongoDB(this.uri); }
}
